@@ -76,7 +76,7 @@ class SimpleSignalRunner(AbstractRunner):
         print('strat sharpe')
         print(sharpe_strat)
         #saver.saveResults(savingKey, hourly_df[['perf_return','turn_over']])
-        saver.saveResults(saving_key, hourly_df['reconstituted_perf'])
+        saver.saveAll(saving_key, hourly_df)
 
 
 
