@@ -6,7 +6,6 @@
 # Built-in packages
 
 # Third party packages
-import numpy as np
 import pandas as pd
 import scipy.cluster.hierarchy as sch
 from scipy.spatial.distance import squareform
@@ -17,7 +16,10 @@ from napoleontoolbox.utility import metrics
 
 from napoleontoolbox.rebalancing import rolling
 
-
+import numpy as np
+import fastcluster
+from scipy.cluster import hierarchy
+from scipy.cluster.hierarchy import fcluster
 # TODO : cython
 
 __all__ = ['minDrawdown','ERC', 'HRP', 'IVP', 'MDP', 'MVP', 'MVP_uc', 'rolling_allocation']
@@ -102,7 +104,6 @@ def ERC(X, w0=None, up_bound=1., low_bound=0.):
 #                     HRP developed by Marcos Lopez de Prado                  #
 # =========================================================================== #
 
-
 def _corr_dist(mat_corr):
     """ Compute a distance matrix based on correlation.
 
@@ -235,29 +236,96 @@ def _get_IVP(mat_cov):
     ivp /= np.sum(ivp)
 
     return ivp
+################################
 
-def HRC(X,method='single', metric='euclidean', low_bound=0., up_bound=1.0):
+def HRC(X, low_bound=0., up_bound=1.0, nb_clusters = 10):
     if not isinstance(X, pd.DataFrame):
         X = pd.DataFrame(X)
 
-    idx = X.columns
     up_bound = max(up_bound, 1 / X.shape[1])
     low_bound = min(low_bound, 1 / X.shape[1])
 
     # Compute covariance and correlation matrix
-    mat_cov = X.cov()
-    mat_corr = X.corr().fillna(0)
-    # Compute distance matrix
-    # print(mat_corr)
-    mat_dist = _corr_dist(mat_corr).fillna(0)
-    mat_dist_corr = squareform(mat_dist)
-    link = sch.linkage(mat_dist_corr, method=method, metric=metric)
-    # Sort linked matrix
-    sortIx = _get_quasi_diag(link)
-    sortIx = mat_corr.index[sortIx].tolist()
-    w = _get_rec_bisec(mat_cov, sortIx)
-    w = w.loc[idx].to_numpy(copy=True).reshape([w.size, 1])
-    return _normalize(w, up_bound=up_bound, low_bound=low_bound)
+    correl_mat = X.corr().fillna(0).values
+
+    dist = 1 - correl_mat
+    dim = len(dist)
+    tri_a, tri_b = np.triu_indices(dim, k=1)
+    Z = fastcluster.linkage(dist[tri_a, tri_b], method='ward')
+
+    permutation = hierarchy.leaves_list(
+        hierarchy.optimal_leaf_ordering(Z, dist[tri_a, tri_b]))
+    ordered_corr = correl_mat[permutation, :][:, permutation]
+
+    clustering_inds = fcluster(Z, nb_clusters, criterion='maxclust')
+    clusters = {i: [] for i in range(min(clustering_inds),
+                                     max(clustering_inds) + 1)}
+    for i, v in enumerate(clustering_inds):
+        clusters[v].append(i)
+    weights = _compute_hrc_allocation(correl_mat, clusters, dim, Z)
+    return _normalize(weights, up_bound=up_bound, low_bound=low_bound)
+
+
+def _seriation(Z, dim, cur_index):
+    if cur_index < dim:
+        return [cur_index]
+    else:
+        left = int(Z[cur_index - dim, 0])
+        right = int(Z[cur_index - dim, 1])
+        return _seriation(Z, dim, left) + _seriation(Z, dim, right)
+
+def _intersection(lst1, lst2):
+    return list(set(lst1) & set(lst2))
+
+def _compute_hrc_allocation(covar, clusters, dim, Z):
+    nb_clusters = len(clusters)
+    assets_weights = np.array([1.] * len(covar))
+    clusters_weights = np.array([1.] * nb_clusters)
+    clusters_var = np.array([0.] * nb_clusters)
+
+    for id_cluster, cluster in clusters.items():
+        cluster_covar = covar[cluster, :][:, cluster]
+        inv_diag = 1 / np.diag(cluster_covar)
+        assets_weights[cluster] = inv_diag / np.sum(inv_diag)
+
+    for id_cluster, cluster in clusters.items():
+        weights = assets_weights[cluster]
+        clusters_var[id_cluster - 1] = np.dot(
+            weights, np.dot(covar[cluster, :][:, cluster], weights))
+
+    for merge in range(nb_clusters - 1):
+        print('id merge:', merge)
+        left = int(Z[dim - 2 - merge, 0])
+        right = int(Z[dim - 2 - merge, 1])
+        left_cluster = _seriation(Z, dim, left)
+        right_cluster = _seriation(Z, dim, right)
+
+        print(len(left_cluster),
+              len(right_cluster))
+        ids_left_cluster = []
+        ids_right_cluster = []
+        for id_cluster, cluster in clusters.items():
+            if sorted(_intersection(left_cluster, cluster)) == sorted(cluster):
+                ids_left_cluster.append(id_cluster)
+            if sorted(_intersection(right_cluster, cluster)) == sorted(cluster):
+                ids_right_cluster.append(id_cluster)
+        ids_left_cluster = np.array(ids_left_cluster) - 1
+        ids_right_cluster = np.array(ids_right_cluster) - 1
+
+        left_cluster_var = np.sum(clusters_var[ids_left_cluster])
+        right_cluster_var = np.sum(clusters_var[ids_right_cluster])
+        alpha = 1 - left_cluster_var / (left_cluster_var + right_cluster_var)
+
+        clusters_weights[ids_left_cluster] = clusters_weights[
+            ids_left_cluster] * alpha
+        clusters_weights[ids_right_cluster] = clusters_weights[
+            ids_right_cluster] * (1 - alpha)
+
+    for id_cluster, cluster in clusters.items():
+        assets_weights[cluster] = assets_weights[cluster] * clusters_weights[
+            id_cluster - 1]
+
+    return assets_weights
 
 
 def HRP(X, method='single', metric='euclidean', low_bound=0., up_bound=1.0):
